// SPDX-License-Identifier: CDDL-1.0
/*
 * This file and its contents are supplied under the terms of the
 * Common Development and Distribution License ("CDDL"), version 1.0.
 * You may only use this file in accordance with the terms of version
 * 1.0 of the CDDL.
 *
 * A full copy of the text of the CDDL should have accompanied this
 * source.  A copy of the CDDL is also available via the Internet at
 * http://www.illumos.org/license/CDDL.
 */

/*
 * Copyright 2020 Toomas Soome <tsoome@me.com>
 */

#ifndef _ZFS_BOOTENV_OS_H
#define	_ZFS_BOOTENV_OS_H

#ifdef __cplusplus
extern "C" {
#endif

#define	BOOTENV_OS		BE_LINUX_VENDOR

#ifdef __cplusplus
}
#endif

#endif /* _ZFS_BOOTENV_OS_H */
